#!/usr/bin/env python2
#
# Usage: ./pyfunge.py [-n] <program> [<time>]
# -n = don't use curses interface (runs much faster)
# <program> = Befunge program to load
# <time> = number of milliseconds to wait between steps
#
# Copyright (c) 2001 Rob Speer
# This program is released under the MIT license. See LICENSE for details.

import traceback, os, string, time
from curses import *
from whrandom import *
from struct import *
from sys import *
debug = 0

ver = 35 # version 0.3.5 (odd minor version = standalone version)

def inrange(x,y):
	if (program.size[1] > y) and (program.size[0] > x) and (y >= 0) and (x >= 0):
		return 1
	else:
		return 0

def chomp(string):
	if string[-1] == '\012': return string[:-1]
	else: return string

def output_text(string):
	global output
	if string != '\000':
		if string >= " " or string == '\012': output += string
		else: output += "?"
	if not curses: stdout.write(string)

class fungerow:
	def __init__(self,x,start):
		self.start = start
		self.length = x
		self.list = [" "]*x
	def __getitem__(self,x):
		a = -self.start + x
		if (a < self.length) and (a >= 0):
			return self.list[a]
		else:
			return " "
	def movestart(self,newstart):
		a = self.start - newstart
		self.list.insert(0, [" "]*a)
		self.start = newstart
		self.size += a
	def resize(self,newlength):
		a = newlength - self.length
		self.list.extend([" "]*a)
		self.length = newlength
	def putlist(self,x,char):
		# This is an absolute position within the list, not adjusted to the
		# start like putgrid
		self.list[x] = char

class fungespace:
	def __init__(self,x,y):
		self.start = (0,0)
		self.size = (x,y)
		self.grid = [fungerow(x,0) for i in range(y)]
	def __getitem__(self,y):
		a = -self.start[1] + y
		if (a < self.size[1]) and (a >= 0):
			return self.grid[y]
		else:
			return fungerow(self.size[0],self.start[0])
	def movestart(self,newstartx,newstarty):
		for row in self.grid:
			row.movestart(newstartx)
		a = self.start[0] - newstartx
		b = self.start[1] - newstarty
		self.grid.insert(0,[fungerow(self.size[0] + a, newstartx) for i in range(b)])
		self.start = (newstartx,newstarty)
		self.size = (self.size[0] + a, self.size[1] + b)
	def resize(self,newx,newy):
		for row in self.grid:
			row.resize(newx)
		b = newy - self.size[1]
		self.grid.extend([fungerow(newx, self.start[0]) for i in range(b)])
		self.size = (newx, newy)
	def putgrid(self,x,y,char):
		if x < self.start[0]:
			self.movestart(x,self.start[1])
		if y < self.start[1]:
			self.movestart(self.start[0],y)
		a = -self.start[0] + x
		b = -self.start[1] + y
		if a >= self.size[0]:
			self.resize(a+1,self.size[1])
		if b >= self.size[1]:
			self.resize(self.size[0],b+1)
		self.grid[b].putlist(a,char)
	def loadat(self,x,y,source):
		width = 0
		for line in source:
			if len(line) > width:
				width = len(line)
		
		for linenum in range(len(source)):
			for charnum in range(width):
				if charnum >= len(source[linenum]):
					self.putgrid(x+charnum,y+linenum,' ')
				else:
					self.putgrid(x+charnum,y+linenum,source[linenum][charnum])



class ip:
	def __init__(self, xcoord, ycoord, xdelta, ydelta):
		global uniqueid
		self.x = xcoord
		self.y = ycoord
		self.dx = xdelta
		self.dy = ydelta
		self.sx = self.sy = 0
		self.stack = [[]]
		uniqueid += 1
		self.id = uniqueid
		self.stringmode = 0
	def setdelta(self, xdelta, ydelta):
		self.dx = xdelta
		self.dy = ydelta
	def setpos(self, xcoord, ycoord):
		self.x = xcoord
		self.y = ycoord
	def setstore(self, xoffset, yoffset):
		self.sx = xoffset
		self.sy = yoffset
	def info(self):
		greatestptx = greatestpty = leastptx = leastpty = 0
		#infolist = [9, # change to 15 when 'i' and 'o' are implemented
		#999999999,     # number of bytes per cell - good approximation to infinity
		#0x50794675,    # Handprint: "PyFu"
		#ver,           # version number
		#1,             # system calls with = go to system()
		#ord(os.sep),   # separator character
		#2,             # 2 dimensions (Befunge)
		#self.id,       # Pointer ID
		#0,             # There are no teams
		infolist = [9, 999999999, 0x50794675, ver, 1, ord(os.sep),\
		2, self.id, 0, self.x, self.y, self.dx, self.dy, self.sx,\
		self.sy, leastptx, leastpty, greatestptx, greatestpty,\
		((time.localtime(0)[0]-1900)*65536 + (time.localtime(0)[1])*256 +\
		time.localtime(0)[2]), ((time.localtime(0)[3])*65536 +\
		(time.localtime(0)[4])*256 + time.localtime(0)[5])]
		for i in argv: # command-line arguments
			for char in i:
				infolist.append(ord(char))
			infolist.append(0)
		infolist.extend([0,0])
		for i in os.environ.keys(): # environment variables
			s = i+'='+os.environ[i]
			for char in s:
				infolist.append(ord(char))
			infolist.append(0)
		infolist.append(0)
		return infolist
	def popstring(self):
		s = ''
		a = self.spop(0)
		while (len(self.stack[0]) > 0) and (a != 0):
			s += chr(a%256)
			a = self.spop(0)
		return s
	def movewrap(self):
		if inrange(self.x + self.dx, self.y + self.dy):
			self.x += self.dx
			self.y += self.dy
		else:
			while inrange(self.x - self.dx, self.y - self.dy):
				self.x -= self.dx
				self.y -= self.dy
	def findnext(self):
		global program
		s = program[self.y][self.x]
		self.movewrap()
		if (self.stringmode == 0) or (s == ' '):
			while program[self.y][self.x] == ' ':
				self.movewrap()
		if (self.stringmode == 0):
			if program[self.y][self.x] == ';':
				self.movewrap()
				while program[self.y][self.x] != ';':
					self.movewrap()
				self.findnext()
	def spop(self,n):
		try:
			a = self.stack[n].pop()
		except:
			a = 0
		return a


def run(c,inst): # c is the current instruction pointer
	global program
	if c.stringmode == 0:
		if inst == '>': # go east
			c.setdelta(1,0)
		elif inst == '<': # go west
			c.setdelta(-1,0)
		elif inst == '^': # go north
			c.setdelta(0,-1)
		elif inst == 'v': # go south
			c.setdelta(0,1)
		elif inst == '?': # go away
			a = randint(0,3)
			b = [(1,0),(-1,0),(0,-1),(0,1)][a]
			c.setdelta(b[0],b[1])
		elif inst == '[': # turn left
			c.setdelta(c.dy, -c.dx)
		elif inst == ']': # turn right
			c.setdelta(-c.dy, c.dx)
		elif inst == 'r': # reverse
			c.setdelta(-c.dx, -c.dy)
		elif inst == 'x': # absolute vector
			b = c.spop(0)
			a = c.spop(0)
			c.setdelta(a, b)
		elif inst == 'z': # nop
			pass
		elif inst == '#': # trampoline
			#c.x += c.dx
			#c.y += c.dy
			c.movewrap()
		elif inst == '@': # die
			if len(iplist) > 1:
				del iplist[iplist.index(c)]
				return
			else:
				exit(0)
		elif inst == 'j': # jump over
			a = c.spop(0)
			#c.x += a * c.dx
			#c.y += a * c.dy
			for i in range(a):
				c.movewrap()
		elif inst == 'q': # quit
			a = c.spop(0)
			exit(a)
		elif inst == 'k': # iterate
			c.findnext()
			newinst = program[c.y][c.x]
			b = c.spop(0)
			for i in range(b):
				run(c,newinst)
		elif inst == '!': # logical not
			a = c.spop(0)
			if a:
				c.stack[0].append(0)
			else:
				c.stack[0].append(1)
		elif inst == '`': # greater than
			b = c.spop(0)
			a = c.spop(0)
			if a > b:
				c.stack[0].append(1)
			else:
				c.stack[0].append(0)
		elif inst == '_': # east-west if
			a = c.spop(0)
			if a:
				c.setdelta(-1,0)
			else:
				c.setdelta(1,0)
		elif inst == '|': # north-south if
			a = c.spop(0)
			if a:
				c.setdelta(0,-1)
			else:
				c.setdelta(0,1)
		elif inst == 'w': # compare and turn
			b = c.spop(0)
			a = c.spop(0)
			if a < b:
				c.setdelta(c.dy, -c.dx)
			elif a > b:
				c.setdelta(-c.dy, c.dx)
		elif ('0' <= inst) and (inst <= '9'): # push 0-9
			c.stack[0].append(int(inst))
		elif inst == 'a': # push ten
			c.stack[0].append(10)
		elif inst == 'b': # push eleven
			c.stack[0].append(11)
		elif inst == 'c': # push twelve
			c.stack[0].append(12)
		elif inst == 'd': # push thirteen
			c.stack[0].append(13)
		elif inst == 'e': # push fourteen
			c.stack[0].append(14)
		elif inst == 'f': # push fifteen
			c.stack[0].append(15)
		elif inst == '+': # add
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].append(long(a+b))
		elif inst == '-': # subtract
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].append(long(b-a))
		elif inst == '*': # multiply
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].append(long(a*b))
		elif inst == '/': # divide
			a = c.spop(0)
			b = c.spop(0)
			if a == 0:
				c.stack[0].append(0)
			else:
				c.stack[0].append(long(b/a))
		elif inst == '%': # modulo
			a = c.spop(0)
			b = c.spop(0)
			if a == 0:
				c.stack[0].append(0)
			else:
				c.stack[0].append(long(b%a))
		elif inst == '"': # start stringmode
			c.stringmode = 1
		elif inst == "'": # fetch character
			#c.x += c.dx
			#c.y += c.dy
			c.movewrap()
			s = ord(program[c.y][c.x])
			c.stack[0].append(s)
		elif inst == "s": # store character
			b = c.y + c.dy
			a = c.x + c.dx
			s = chr(c.spop(0) % 256)
			program.putgrid(a,b,s)
			if (0 <= b < 25) and (0 <= a < 80) and curses:
				mungechar(b,a,s,A_BOLD)
			
		elif inst == "$": # pop
			if len(c.stack[0]) > 0: del c.stack[0][-1]
		elif inst == ":": # duplicate
			a = c.spop(0)
			c.stack[0].extend([a,a])
		elif inst == "\\": # swap
			a = c.spop(0)
			b = c.spop(0)
			c.stack[0].extend([a,b])
		elif inst == "n": # stack clear
			c.stack[0] = []
		elif inst == "{": # begin block
			n = c.spop(0)
			c.stack.insert(0,[])
			if n > 0:
				if len(c.stack[1]) < n:
					c.stack[0] = c.stack[1] + [0]*(n - len(c.stack[1]))
					c.stack[1] = []
				else:
					c.stack[0] = c.stack[1][-n:]
					c.stack[1][-n:] = []
			elif n < 0:
				c.stack[1].extend([0]*abs(n))
			c.stack[1].extend([c.sx,c.sy])
			c.setstore(c.x + c.dx, c.y + c.dy)
		elif inst == "}": # end block
			if len(c.stack) < 2:
				c.setdelta(-c.dx,-c.dy) # reflect on failure
			else:
				n = c.spop(0)
				b = c.spop(1)
				a = c.spop(1)
				c.setstore(a,b)
				if n > 0:
					if len(c.stack[0]) < n:
						c.stack[1].extend(c.stack[0])
						c.stack[1].extend([0]*(n - len(c.stack[0])))
					else:
						c.stack[1].extend(c.stack[0][-n:])
				elif n < 0:
					del c.stack[1][n:]
				del c.stack[0]
		elif inst == "u": # stack under stack
			if len(c.stack) < 2:
				c.setdelta(-c.dx,-c.dy) # reflect on failure
			else:
				n = c.spop(0)
				if n > 0:
					for i in range(n):
						c.stack[0].append(c.spop(1))
				elif n < 0:
					for i in range(abs(n)):
						c.stack[1].append(c.spop(0))
		elif inst == "g": # get
			b = c.spop(0) + c.sy
			a = c.spop(0) + c.sx
			if program.size[1] > b:
				if program.size[0] > a:
					s = ord(program[b][a])
					c.stack[0].append(s)
				else:
					c.stack[0].append(32)
			else:
				c.stack[0].append(32)
		elif inst == "p": # put
			b = c.spop(0) + c.sy
			a = c.spop(0) + c.sx
			s = chr(c.spop(0) % 256)
			program.putgrid(a,b,s)
			if (0 <= b < 25) and (0 <= a < 80) and curses:
				mungechar(b,a,s,A_BOLD)
		elif inst == ".": # output decimal
			if len(c.stack[0]) == 0:
				c.setdelta(-c.dx,-c.dy)
			else:
				output_text(str(c.spop(0))+" ")
		elif inst == ",": # output character
			s = chr(c.spop(0) % 256)
			output_text(s)
		elif inst == "&": # input decimal
			s = ''
			char = stdin.read(1)
			while char not in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '']:
				char = stdin.read(1)
			if (char == ''):
				c.setdelta(-c.dx,-c.dy)
			else:
				while char in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
					output_text(char)
					if curses:
						stdscr.addstr(33, 0,\
						"\n".join(output[-239:].splitlines()[-3:])+"\n",\
						color_pair(3) | A_BOLD)
						stdscr.refresh()
					s += char
					char = stdin.read(1)
				output_text("\n")
				c.stack[0].append(int(s))
		elif inst == "~": # input character
			char = stdin.read(1)
			if (char == ''):
				c.setdelta(-c.dx,-c.dy)
			else:
				a = ord(char)
				if a == 13: a = 10
				c.stack[0].append(a)
		elif inst == "=": # system execute
			s = c.popstring()
			c.stack[0].append(os.system(s))
		elif inst == "(": # load fingerprint
			c.setdelta(-c.dx,-c.dy) # fingerprints are not implemented
		elif inst == ")": # unload fingerprint
			c.setdelta(-c.dx,-c.dy) # fingerprints are not implemented
		elif inst == "t": # split
			n = iplist.index(c)
			iplist.insert(n, ip(c.x,c.y,-c.dx,-c.dy))
		elif inst == "y": # get info
			n = c.spop(0)
			temp = c.info()
			if n > 0:
				if n >= len(temp):
					c.stack[0].append(c.stack[0][n-len(temp)])
				else:
					c.stack[0].append(c.info()[n-1])
			else:
				temp = c.info()
				temp.reverse()
				c.stack[0].extend(temp)
			del temp
		else:
			a = ord(inst)
			if debug: stderr.write("Unimplemented instruction: "+`a`+" on IP\
#"+`c.id`+", at "+`(c.x,c.y)`+".\n")
			#c.setdelta(-c.dx,-c.dy)
	else: # string mode
		if inst == '"': # end stringmode
			c.stringmode = 0
		else:
			s = ord(inst)
			c.stack[0].append(s)

def mungechar(y, x, char, attribs=0):
	if (ord(char) < 0) or (ord(char) > 254): stdscr.addch(y, x, "~",\
	color_pair(4) | A_BOLD | attribs)
	elif ord(char) == 127: stdscr.addch(y, x, "?", color_pair(4) | A_BOLD | attribs)
	elif ord(char) >= 32: stdscr.addch(y, x, char, attribs)
	else: stdscr.addch(y, x, ord(char)+64, color_pair(4) | A_BOLD | attribs)

def redraw():
	if curses:
		for y in range(25):
			for x in range(80):
				mungechar(y, x, program[y][x])
#		for p in iplist:
#			mungechar(p.y, p.x, program[p.y][p.x], A_REVERSE)

def showstack(stack):
	out = ""
	for i in stack:
		if 32 <= i < 128:
			out += chr(i)+" "
		else:
			out += "["+str(i)+"] "
	return out


def step(c,inst):
	global toclear
	
	if delay: time.sleep(delay)
	run(c,inst)
	if curses:
		for i in toclear:
			mungechar(i[0], i[1], program[i[0]][i[1]])
		toclear = []
		for p in iplist:
			mungechar(p.y, p.x, program[p.y][p.x], A_REVERSE)
			toclear += [(p.y, p.x)]
		
		stdscr.addstr(33, 0, "\n".join(output[-239:].splitlines()[-3:])+"\n",\
		color_pair(3) | A_BOLD)
		stdscr.addstr(30, 0, showstack(iplist[0].stack[0])[:239]+'\n',\
		color_pair(2))
		stdscr.refresh()
	c.findnext()

def main(stdscr):
	global output, ticks, uniqueid, program, iplist, toclear, delay
	if curses:
		# Define the color pairs
		init_pair(1, COLOR_RED, COLOR_BLACK)
		init_pair(2, COLOR_GREEN, COLOR_BLACK)
		init_pair(3, COLOR_YELLOW, COLOR_BLACK)
		init_pair(4, COLOR_BLUE, COLOR_BLACK)
	
		curs_set(0) # make cursor invisible
	
	output = ""
	ticks = 0
	uniqueid = 0
	if len(argv) > 2: delay = int(argv[2])/1000.0
	elif curses: delay = 0.10
	else: delay = 0
	
	if os.access(argv[1],os.R_OK) == 0:
		exit("Can't read " + argv[1])
	file = open(argv[1])
	source = map(chomp,file.readlines())
	file.close()

	program = fungespace(80,25)
	program.loadat(0,0,source)

	iplist = [ip(program.size[0]-1,0,1,0)]
	iplist[0].findnext()
	
	toclear = []
	redraw()
	
	while 1: # main loop
		ticks += 1
		for each_ip in iplist:
			instruct = program[each_ip.y][each_ip.x]
			step(each_ip,instruct)


if __name__=='__main__':
	if "-n" in argv:
		del argv[argv.index("-n")]
		curses = 0
		try: main(None)
		except SystemExit, detail:
			print detail
		except:
			traceback.print_exc()
	else:
		curses = 1
		try:
			# Initialize curses
			stdscr=initscr()
			start_color()
			noecho() ; cbreak() 
			stdscr.keypad(1)
			
			main(stdscr)  # Enter the main loop
			
			stdscr.keypad(0)
			echo() ; nocbreak()
			endwin()  # Terminate curses
		except SystemExit, detail:
			# Restore the terminal to a sane state
			stdscr.keypad(0)
			echo() ; nocbreak()
			endwin()
			print detail
		except:
			# Restore the terminal to a sane state
			stdscr.keypad(0)
			echo() ; nocbreak()
			endwin() 
			traceback.print_exc()

